// ==============================================================
//
// f9_v12_stg2
// Control module for f9_v12_stg2 vessel class
//
// ==============================================================

#define STRICT 1
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>
#include "DlgCtrl.h"
#include "f9_v12_stg2_resource.h"
#include "gcAPI.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

#define CTDN	1
#define SEP		2
#define SEPST	3
#define SEPFR	4


const DWORD COMM_SEND_STR = 2300;
const DWORD COMM_SEND_AFS = 2301;
const DWORD COMM_CALC_MASS = 2302;
const DWORD COMM_CALC_FRE = 2303;
const DWORD COMM_SEND_EDLO = 2304;
const double STG1ISP0 = 2767;
const double STG1ISPV = 3051;
const double STG1MAINTH0 = 756222;//654000;
const double STG1MAINTH1 = 825111;//741000;
const double STG1RCSTH = 38000;
const double STG1DRYMASS = 22500;//25600;//19350;
const double STG1FUELMASS = 402500;//384000;

const double STG2ISP0 = 3151;//3051
const double STG2ISPV = 3549; // ???
const double STG2MAINTH0 = 825111;//741000;
const double STG2MAINTH1 = 934000;//801000;
const double STG2RCSTH1 = 500;
const double STG2RCSTH2 = 13000;
const double STG2FUELMASS = 104275;//93100;
const double STG2DRYMASS = 4300;//5800;
const double STG2RCSFUELMASS = 100;
const double STG2RCSISP = 800;

const VECTOR3 STG1_OFF = _V(0,0,-10);
const VECTOR3 STG1INT_OFF = _V(0,0,-10);
const VECTOR3 STG1FAIR_OFF = _V(0,0,31.384);
const VECTOR3 STG1_FAIR1_OFF = _V(0,-1,31.384);
const VECTOR3 STG1_FAIR2_OFF = _V(0, 1,31.384);
const VECTOR3 STG1STG2_OFF = _V(0,0,19.596);
const VECTOR3 STG2FAIR_OFF = _V(0,0,12.288);
const VECTOR3 STG2_FAIR1_OFF = _V(0,-1,12.288);
const VECTOR3 STG2_FAIR2_OFF = _V(0, 1,12.288);

const double  FAIR_MASS = 1750;

const VECTOR3 ADAPTOR_OFF = _V(0,0,26.646);

//const VECTOR3 FAIR_OFF = _V(0,0,30.479);

char refname[256];
char pldname[256];
int afs_status = 0;
double tgt_inc = 0;
double min_inc = 0;
double fuel_reserve = 0;
double pld_mass = 0;
int dlg_status = 0;
int dlgl_status = 0;
int pld_att_status = 0;
int calc_sw = 0;
int intfair_status = 0;
int eng_status = 0;
int apop_status = 0;
int edlo_sw = 0;
int nolegs = 0;
double ap_tgt = 135000;
const double mu = 398600439968871.2;
const double rt = 6371010;
double op_meco_alt = 180000;
double op_tgt = 200000;
double fuel_res_est = 0;
ATTACHMENTHANDLE pld_attach;
ATTACHMENTHANDLE lpd_attach;
OBJHANDLE hMe;
OBJHANDLE hGref;
OBJHANDLE bchk;
double asds_core_lat = 0;
double asds_core_lng = 0;
double frv_core_lat = 0;
double frv_core_lng = 0;
NOTEHANDLE hNote;
int lot_hh = 0;
int lot_mm = 0;
int lot_ss = 0;
int lot_status = 0;
double lot_timer = 0;

static TOUCHDOWNVTX tdvtx_geardown[4] = {
	{_V(    0,-5, -43), 1e6, 1e5, 0.8, 0.8},
	{_V(-5, 3, -43), 1e6, 1e5, 0.8, 0.8},
	{_V( 5, 3,-43), 1e6, 1e5, 0.8, 0.8},
	{_V( 0, 0 , 40), 1e6, 1e5, 1.5}
};

static TOUCHDOWNVTX tdvtx_gearup[4] = {
	{_V(    0,-5, -10), 1e6, 1e5, 0.8, 0.8},
	{_V(-5, 3, -10), 1e6, 1e5, 0.8, 0.8},
	{_V( 5, 3,-10), 1e6, 1e5, 0.8, 0.8},
	{_V( 0, 0 , 10), 1e6, 1e5, 1.5}
};

#define GREEN RGB(125, 255, 125)

typedef struct {
	HINSTANCE hDLL;
} GDIParams;
GDIParams g_Param;
//#define LOADBMP(id) (LoadBitmap (g_Param.hDLL, MAKEINTRESOURCE (id)))


void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0};


	aoa += PI;
	int idx = max (0, min (15, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	*cd = 0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2);
	
	
}

// 2. horizontal lift component (vertical stabiliser and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0};

	beta += PI;
	int idx = max (0, min (15, (int)(beta*istep)));
	double d = beta*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	*cd = 0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2);
	
}

// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class f9_v12_stg2: public VESSEL3 {
public:
	f9_v12_stg2 (OBJHANDLE hVessel, int flightmodel);
	~f9_v12_stg2 ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPostCreation();
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int MyID;	// this is your unique Id, it will identify your ship in OrbiterSound.
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	bool clbkLoadGenericCockpit ();
	
private:

	char adname [256];

	THRUSTER_HANDLE th_main[2],th_rcs[24], th_group[9], th_dump;
	THGROUP_HANDLE htgrpmain;
	PROPELLANT_HANDLE hProp1,hProp2, hRCSProp2;
	THGROUP_HANDLE htgrppup2,htgrppdown2,htgrpbright2, htgrpbleft2, htgrpyright2, htgrpyleft2;

	ELEMENTS el;
	ORBITPARAM prm;

	OBJHANDLE hpd;
	OBJHANDLE bbchk;
	PointLight *spt_light;

	SURFHANDLE bs1, bs2;

	void JettisonPld(void);
	void JettisonFair(void);
	void JettisonStg(void);
	void ConfigMain(void);
	void ConfigRCS2(void);
	void RevertRCS2(void);
	void CalcEmptyMass(void);
	void DoAscent(void);
	void DoOrbit(void);
	double CalcAzimuth();
	void CalcFlightVector();
	void GoToFlightVector();
	double FinalV();
	double GetOS();
	void RollToAz();
	void SetCamDir();
	void KillAllRCS2();
	//void InitDialog(HWND hWnd);
	void CalcFuelRes();
	void DoLOT();
	void CalcAsdsPos();
	

	double step_time, old_ap_dist, pitch_angle, bias, timer, vrot_lim, pst_lvl, pst_lvl2;
	int op_status, ap_status, rcs_sw2;
	int ap_function, cam_status, drg, nfr;
	int start_sw, tu_sw, skin_sw;
	int disable_fair_rec;

	double azimuth;	
	VECTOR3 azim_local, vazim1, vazim2, flight_vector;

	double atl_current;
	
	
};

f9_v12_stg2::f9_v12_stg2 (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

f9_v12_stg2::~f9_v12_stg2 ()
{
	

}


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void f9_v12_stg2::clbkSetClassCaps (FILEHANDLE cfg)
{
	strcpy(refname, "NULL");
	strcpy(pldname, "NULL");

	step_time = 0;
	ap_status = 0;
	op_status = 0;
	cam_status = 0;
	ap_function = 0;
	old_ap_dist = 0;
	bias = 0;
	pitch_angle = 90*RAD;
	//intfair_status = 0;
	drg = 0;
	nfr = 0;
	//eng_status = 0;
	//calc_sw = 0;
	rcs_sw2 = 0;
	vrot_lim = 5*RAD;
	azimuth = 0;
	azim_local = _V(0,-1,0);
	vazim1 = _V(0,1,0);
	vazim2 = _V(0,1,0);
	flight_vector = _V(0,1,0);
	timer = 0;
	start_sw = 0;
	pst_lvl = 0;
	pst_lvl2 = 0;
	tu_sw = 0;
	atl_current = 1;
	skin_sw = 0;
	disable_fair_rec = 0;

	gcInitialize();

	// physical vessel parameters
	SetSize (30);
	SetEmptyMass (STG1DRYMASS+STG2DRYMASS+FAIR_MASS);
	SetPMI(_V(129,129,1.9));
	SetCrossSections (_V(159,159,16));
	SetCW(0.5,1,1,1);
	SetRotDrag (_V(0.5,0.5,0.1));
	SetTouchdownPoints ( tdvtx_geardown, 4);

	// camera parameters
	SetCameraOffset (_V(0,3,0));
	SetCameraRotationRange(PI,PI,(0.4*PI),(0.4*PI));

	oapiReadItem_int (cfg, "NoFairRecovery", disable_fair_rec);

	oapiReadItem_int (cfg, "NoFairing", nfr);
	
	if(oapiReadItem_int (cfg, "Dragon", drg) == true)
	{
		if (drg == 1)
		{
		SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_stg2_drg", &STG1STG2_OFF),MESHVIS_ALWAYS);
		intfair_status = 1;
		}
		else if (drg == 2)
		{
		SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_stg2_drg2", &STG1STG2_OFF),MESHVIS_ALWAYS);
		intfair_status = 1;
		}
		else SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_stg2", &STG1STG2_OFF),MESHVIS_ALWAYS);
	}
	else SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_stg2", &STG1STG2_OFF),MESHVIS_ALWAYS);

	MESHHANDLE hMhull;

	if(oapiReadItem_int (cfg, "Expendable", nolegs) == true)
	{
	hMhull = oapiLoadMeshGlobal("Falcon9_BLOCK4/f9_v12_stg1_nolegs");
	SetMeshVisibilityMode(AddMesh(hMhull, &STG1_OFF),MESHVIS_ALWAYS);
	SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_interstg_nolegs", &STG1INT_OFF),MESHVIS_ALWAYS);
	nolegs = 1;
	}
	else 
	{
	hMhull = oapiLoadMeshGlobal("Falcon9_BLOCK4/f9_v12_stg1");
	SetMeshVisibilityMode(AddMesh(hMhull, &STG1_OFF),MESHVIS_ALWAYS);
	SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_interstg", &STG1INT_OFF),MESHVIS_ALWAYS);
	nolegs = 0;
	}
	
	SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_fair_r", &STG1FAIR_OFF),MESHVIS_ALWAYS);
	SetMeshVisibilityMode(AddMesh("Falcon9_BLOCK4/f9_v12_fair_l", &STG1FAIR_OFF),MESHVIS_ALWAYS);

	if(oapiReadItem_string(cfg, "Adaptor", adname) == TRUE)
	{
	SetMeshVisibilityMode(AddMesh(adname, &ADAPTOR_OFF),MESHVIS_ALWAYS);
	}

	/////// charred skin
	if(oapiReadItem_int (cfg, "Charred", skin_sw) == true)
	{
	if(skin_sw == 1)
		{
		bs2 = oapiLoadTexture ("Falcon9_BLOCK4/f9_hull_logo_char.dds", true);
		bs1 = oapiGetTextureHandle (hMhull, 1);
		oapiBlt(bs1, bs2, 0, 0, 0, 0, 256, 1024);
		}
	}

	// propellant resources
	
	hProp1 = CreatePropellantResource (STG1FUELMASS,STG1FUELMASS);
	hProp2 = CreatePropellantResource (STG2FUELMASS,STG2FUELMASS);
	hRCSProp2 = CreatePropellantResource (STG2RCSFUELMASS,STG2RCSFUELMASS);

	// ***************** thruster definitions *******************
	

	th_rcs[0] = CreateThruster (_V( 0, 2, -31), _V( 1,0,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);
	th_rcs[1] = CreateThruster (_V( 0, 2, -31), _V(-1,0,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);
	th_rcs[2] = CreateThruster (_V( 0,-2, -31), _V( 1,0,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);
	th_rcs[3] = CreateThruster (_V( 0,-2, -31), _V(-1,0,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);

	th_rcs[4] = CreateThruster (_V( 0, 4, -31), _V( 1,0,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);
	th_rcs[5] = CreateThruster (_V( 0, 4, -31), _V(-1,0,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);
	th_rcs[6] = CreateThruster (_V( 0,-4, -31), _V( 1,0,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);
	th_rcs[7] = CreateThruster (_V( 0,-4, -31), _V(-1,0,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);

	th_rcs[8] = CreateThruster (_V( 2, 0, -31), _V(0, 1,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);
	th_rcs[9] = CreateThruster (_V( 2, 0, -31), _V(0,-1,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);
	th_rcs[10] = CreateThruster (_V(-2,0, -31), _V(0, 1,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);
	th_rcs[11] = CreateThruster (_V(-2,0, -31), _V(0,-1,0), STG1RCSTH, hProp1, STG1ISPV, STG1ISP0);


	////// stg2

	th_rcs[12] = CreateThruster (_V( 0, 0, -9.8), _V( 1,0,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); //yl
	th_rcs[13] = CreateThruster (_V( 0, 0, -9.8), _V(-1,0,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); //  yr
	th_rcs[14] = CreateThruster (_V( 0, 0, -9.8), _V( 0,-1,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); // pup
	th_rcs[15] = CreateThruster (_V( 0, 0, -9.8), _V( 0,1,0), STG2RCSTH2, hProp2, STG2ISPV, STG2ISP0); // pd

	th_rcs[16] = CreateThruster (_V(1.706,0.123,-4.365), _V(-0.25,-0.866,0.433), STG2RCSTH1, hRCSProp2, STG2RCSISP, STG2RCSISP); // br
	th_rcs[17] = CreateThruster (_V(-1.706,0.123,-4.365), _V( 0.25,-0.866,0.433), STG2RCSTH1, hRCSProp2, STG2RCSISP, STG2RCSISP);// bl
	th_rcs[18] = CreateThruster (_V(1.706,-0.123,-4.365), _V(-0.25,0.866,0.433), STG2RCSTH1, hRCSProp2, STG2RCSISP, STG2RCSISP); // bl
	th_rcs[19] = CreateThruster (_V(-1.706,-0.123,-4.365), _V( 0.25,0.866,0.433), STG2RCSTH1, hRCSProp2, STG2RCSISP, STG2RCSISP);// br

	th_rcs[20] = CreateThruster (_V(0.123,1.706,-4.365), _V(-0.866,-0.25,0.433), STG2RCSTH1, hRCSProp2, STG2RCSISP, STG2RCSISP); // yl
	th_rcs[21] = CreateThruster (_V(-0.123,1.706,-4.365), _V( 0.866,-0.25,0.433), STG2RCSTH1, hRCSProp2, STG2RCSISP, STG2RCSISP);// yr
	th_rcs[22] = CreateThruster (_V(0.123,-1.706,-4.365), _V(-0.866,0.25,0.433), STG2RCSTH1, hRCSProp2, STG2RCSISP, STG2RCSISP); // yl
	th_rcs[23] = CreateThruster (_V(-0.123,-1.706,-4.365), _V( 0.866,0.25,0.433), STG2RCSTH1, hRCSProp2, STG2RCSISP, STG2RCSISP);// yr

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[11];
	htgrppup2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[10];
	htgrppdown2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	htgrpyright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	htgrpyleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[4];
	th_group[1] = th_rcs[7];
	htgrpbright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[5];
	th_group[1] = th_rcs[6];
	htgrpbleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);


	th_main[0] = CreateThruster (_V( 0, 0, -31), _V( 0,0,1), 9*STG1MAINTH1, hProp1, STG1ISPV, STG1ISP0);
	th_main[1] = CreateThruster (_V( 0, 0, -9.8), _V( 0,0,1), STG2MAINTH1, hProp2, STG2ISPV, STG2ISP0);
	th_dump = CreateThruster (_V( 0, 0, -9.8), _V( 0,0,1), 4000, hProp2, 100, 100);

	th_group[0] = th_main[0];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);

	SURFHANDLE exhaust = oapiRegisterExhaustTexture ("Exhaust2");
	SURFHANDLE exhaustrcs = oapiRegisterExhaustTexture ("exhaust_atrcs");


	double exlen = 30;
	double exwid = 1;

	AddExhaust(th_main[0],exlen,exwid,_V(0,0,-31),_V(0,0,-1),exhaust);
	AddExhaust(th_main[0],exlen,exwid,_V(1.32,0,-31),_V(0,0,-1),exhaust);
	AddExhaust(th_main[0],exlen,exwid,_V(-1.32,0,-31),_V(0,0,-1),exhaust);
	AddExhaust(th_main[0],exlen,exwid,_V(0,1.32,-31),_V(0,0,-1),exhaust);
	AddExhaust(th_main[0],exlen,exwid,_V(0,-1.32,-31),_V(0,0,-1),exhaust);
	AddExhaust(th_main[0],exlen,exwid,_V(0.931,0.931,-31),_V(0,0,-1),exhaust);
	AddExhaust(th_main[0],exlen,exwid,_V(-0.931,-0.931,-31),_V(0,0,-1),exhaust);
	AddExhaust(th_main[0],exlen,exwid,_V(-0.931,0.931,-31),_V(0,0,-1),exhaust);
	AddExhaust(th_main[0],exlen,exwid,_V(0.931,-0.931,-31),_V(0,0,-1),exhaust);

	AddExhaust(th_main[1],20,2,_V(0,0,-10),_V(0,0,-1),exhaust);

	double exlenrcs = 2;
	double exwidrcs = 0.2;

	AddExhaust(th_rcs[16],exlenrcs,exwidrcs,exhaustrcs);
	AddExhaust(th_rcs[17],exlenrcs,exwidrcs,exhaustrcs);
	AddExhaust(th_rcs[18],exlenrcs,exwidrcs,exhaustrcs);
	AddExhaust(th_rcs[19],exlenrcs,exwidrcs,exhaustrcs);

	AddExhaust(th_rcs[20],exlenrcs,exwidrcs,exhaustrcs);
	AddExhaust(th_rcs[21],exlenrcs,exwidrcs,exhaustrcs);
	AddExhaust(th_rcs[22],exlenrcs,exwidrcs,exhaustrcs);
	AddExhaust(th_rcs[23],exlenrcs,exwidrcs,exhaustrcs);

	////// p streams
	SURFHANDLE exhaust2 = oapiRegisterParticleTexture ("Contrail3");

	PARTICLESTREAMSPEC exhaust_main1 = {
		0, 3, 40, 150, 0.01, 0.15, 1, 2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		exhaust2
	};


	PARTICLESTREAMSPEC exhaust_main2 = {
		0, 5, 40, 150, 0.07, 0.075, -20, 2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 0.7e-7, 1,
		exhaust2
	};

	///max q pstream
	SURFHANDLE exhaust3 = oapiRegisterParticleTexture ("Contrail1");
	PARTICLESTREAMSPEC maxqtrail = {
		0, 3, 20, 50, 0.03, 0.1, 1, 3, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		exhaust3
	};
	AddParticleStream(&maxqtrail,_V(0,0,25),_V(0,0,-1), &pst_lvl);

	////// condensation
	PARTICLESTREAMSPEC condensation = {
		0, 2.0, 4, 0.5, 0.25, 4, 0.0, 0.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0.5, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		exhaust3
	};

	///// fuel dump
	PARTICLESTREAMSPEC fuel_dump = {
		0, 3.0, 4, 150, 0.5, 6, 20.0, 10.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0.5, 0.5,
		PARTICLESTREAMSPEC::ATM_FLAT, 0.5, 0.5,
		exhaust3
	};

	if (gcEnabled() == false) AddParticleStream(&condensation,_V(0,0,8),_V(0,0,-1), &pst_lvl2);

	if (gcEnabled() == false) AddParticleStream(&condensation,_V(0,0,24),_V(0,0,-1), &pst_lvl2);

	//AddParticleStream(&condensation,_V(0,0,8),_V(0,0,-1), &pst_lvl2);
	//AddParticleStream(&condensation,_V(0,0,24),_V(0,0,-1), &pst_lvl2);
	
	//////////

	AddExhaustStream(th_main[0],_V(0,0,-31),&exhaust_main1);
	AddExhaustStream(th_main[0],_V(1.32,0,-31),&exhaust_main1);
	AddExhaustStream(th_main[0],_V(-1.32,0,-31),&exhaust_main1);
	AddExhaustStream(th_main[0],_V(0,1.32,-31),&exhaust_main1);
	AddExhaustStream(th_main[0],_V(0,-1.32,-31),&exhaust_main1);
	AddExhaustStream(th_main[0],_V(0.931,0.931,-31),&exhaust_main1);
	AddExhaustStream(th_main[0],_V(-0.931,-0.931,-31),&exhaust_main1);
	AddExhaustStream(th_main[0],_V(-0.931,0.931,-31),&exhaust_main1);
	AddExhaustStream(th_main[0],_V(0.931,-0.931,-31),&exhaust_main1);

	AddExhaustStream(th_main[0],_V(0,0,-31),&exhaust_main2);
	AddExhaustStream(th_main[0],_V(1.32,0,-31),&exhaust_main2);
	AddExhaustStream(th_main[0],_V(-1.32,0,-31),&exhaust_main2);
	AddExhaustStream(th_main[0],_V(0,1.32,-31),&exhaust_main2);
	AddExhaustStream(th_main[0],_V(0,-1.32,-31),&exhaust_main2);
	AddExhaustStream(th_main[0],_V(0.931,0.931,-31),&exhaust_main2);
	AddExhaustStream(th_main[0],_V(-0.931,-0.931,-31),&exhaust_main2);
	AddExhaustStream(th_main[0],_V(-0.931,0.931,-31),&exhaust_main2);
	AddExhaustStream(th_main[0],_V(0.931,-0.931,-31),&exhaust_main2);

	AddExhaustStream(th_dump, _V(0,0, -9.8), &fuel_dump);
	///////////////////////////////////////////////////////
	/////AirFoils

	CreateAirfoil (LIFT_VERTICAL, _V(0,0,0), VLiftCoeff, 25, 100, 0.16);
	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeff, 25, 100, 0.16);

	////////

	pld_attach = CreateAttachment (false, _V( 0, 0, 26.646), _V( 0,0,1), _V(0,1,0), "F9");
	lpd_attach = CreateAttachment (true, _V( 0, 0, -29.45), _V( 0,0,-1), _V(0,1,0), "LP");

	//////////
	//local light
	COLOUR4 col_d = {0.125,0.1,0.06125,0};
	COLOUR4 col_s = {0,0,0,0};
	COLOUR4 col_a = {0,0,0,0};
	COLOUR4 col_white = {1,1,1,0};
	//spt_light = (SpotLight*)AddSpotLight(_V(0,0,-21), _V(0,0,-1),5000, 1e-1, 0, 1e-7, RAD*170, RAD*180, col_d, col_s, col_a);
	spt_light = (PointLight*)AddPointLight(_V(0,0,-27),5000, 1e-1, 0, 1e-6, col_d, col_s, col_a);
	//pnt_light->Activate(false);
	spt_light->Activate(false);

}

void f9_v12_stg2::clbkPostCreation ()
{
 hGref = GetGravityRef();
 hMe = GetHandle();
 OBJHANDLE bbtgt = oapiGetObjectByName(refname);
 if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, refname);
 if (bbtgt == NULL)
 {
	strcpy(refname, "NULL");
 }

 if (drg == 1)intfair_status = 1;
 if (nfr == 1)intfair_status = 1;
 
 if(intfair_status == 1)
  {
	if (eng_status == 1)
	 {
	 DelMesh(4);
	 DelMesh(3);
	 DelMesh(2);
	 DelMesh(1);
	 //double thmain_lvl = GetThrusterLevel(th_main[1]);
	 ConfigMain();
	 //SetThrusterLevel(th_main[1], thmain_lvl);
	 }
	else
	 {	
	 DelMesh(4);
	 DelMesh(3);
	 }
  }
  else if (eng_status == 1)
  {
	  DelMesh(2);
	  DelMesh(1);
	  //double thmain_lvl = GetThrusterLevel(th_main[1]);
	  ConfigMain();
	  //SetThrusterLevel(th_main[1], thmain_lvl);
  }

  CalcEmptyMass();
  CalcFuelRes();
 
	double lat, lng, rds;
	GetEquPos(lng,lat,rds);
	min_inc = lat;
	if (abs(tgt_inc) < abs(lat)) tgt_inc = lat;

	///////////// sound
	MyID=ConnectToOrbiterSoundDLL(GetHandle());

	RequestLoadVesselWave(MyID,CTDN,"Sound\\Vessel\\countdown1.wav",DEFAULT);
	RequestLoadVesselWave(MyID,SEP,"Sound\\Vessel\\undocking.wav",BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave(MyID,SEPST,"Sound\\_CustomVesselsSounds\\Falcon9_BLOCK4\\stgsep.wav",DEFAULT);
	RequestLoadVesselWave(MyID,SEPFR,"Sound\\_CustomVesselsSounds\\Falcon9_BLOCK4\\frsep.wav",DEFAULT);
	
	SoundOptionOnOff(MyID,PLAYRADIOATC,FALSE);
	//SoundOptionOnOff3(MyID,PLAYWHENATTITUDEMODECHANGE,FALSE);
	//SoundOptionOnOff3(MyID,PLAYGPWS,FALSE);
	SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	//SoundOptionOnOff3(MyID,PLAYRADARBIP,FALSE);

}


// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new f9_v12_stg2 (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (f9_v12_stg2*)vessel;
}


void f9_v12_stg2::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	for (int i = 0; i < (int)strlen(refname); i++)
	{
		if (refname[i] == ' ')
		{
			refname[i] = '+';
			break;
		}
	}

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	//sprintf (cbuf, "%d %0.4f",);
	//oapiWriteScenario_string (scn, "LEGS", cbuf);

	sprintf (cbuf, "%d",intfair_status);
	oapiWriteScenario_string (scn, "FAIRS", cbuf);

	sprintf (cbuf, "%d %d %d %d %d %d %d", eng_status, ap_status, ap_function, op_status, afs_status, start_sw, edlo_sw);
	oapiWriteScenario_string (scn, "SOST", cbuf);

	sprintf (cbuf, "%0.4f %0.1f %0.1f %0.1f %0.1f %s %0.4f", tgt_inc, op_meco_alt, op_tgt, fuel_reserve, ap_tgt, refname, pitch_angle);
	oapiWriteScenario_string (scn, "SOPR", cbuf);
	lot_hh, lot_mm, lot_ss, lot_status;

	sprintf (cbuf, "%d %d %d %d", lot_hh, lot_mm, lot_ss, lot_status);
	oapiWriteScenario_string (scn, "LOTM", cbuf);
}
void f9_v12_stg2::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "FAIRS", 5)) {
		sscanf (line+5, "%d", &intfair_status);
		} else if (!strnicmp (line, "SOST", 4)) {
		sscanf (line+4, "%d%d%d%d%d%d%d", &eng_status, &ap_status, &ap_function, &op_status, &afs_status, &start_sw, &edlo_sw);
		} else if (!strnicmp (line, "SOPR", 4)) {
			sscanf (line+4, "%lf%lf%lf%lf%lf%s%lf",&tgt_inc, &op_meco_alt, &op_tgt, &fuel_reserve, &ap_tgt, &refname, &pitch_angle);
		} else if (!strnicmp (line, "LOTM", 4)) {
			sscanf (line+4, "%d%d%d%d",&lot_hh, &lot_mm, &lot_ss, &lot_status);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }


	for (int i = 0; i < (int)strlen(refname); i++)
	{
		if (refname[i] == '+')
		{
			refname[i] = ' ';
			break;
		}
	}

	//CalcFuelRes();
}

bool SetINC(void *id, char *str, void *data)
{
	float new_inc;
	sscanf(str, "%f", &new_inc);
	if (abs(new_inc) < abs(min_inc*DEG))return false;
	else 
		{
		tgt_inc = new_inc*RAD;
		return true;
		}
}

bool SetRF(void *id, char *str, void *data)
{
	float new_rf;
	sscanf(str, "%f", &new_rf);
	if (new_rf < 0)return false;
	else if (new_rf > 80000)return false;
	else 
		{
		fuel_reserve = new_rf;
		if (ap_tgt > (450000 - (3*fuel_reserve)))ap_tgt = (450000 - (3*fuel_reserve));
		if (ap_tgt < 100000)ap_tgt = 100000;
		return true;
		}
}

bool SetAP(void *id, char *str, void *data)
{
	float new_ap;
	sscanf(str, "%f", &new_ap);
	if (new_ap < 100)return false;
	else if (new_ap > 450)return false;
	else 
		{
		ap_tgt = new_ap*1000;
		if (fuel_reserve > ((450000 - ap_tgt)/3)) fuel_reserve = ((450000 - ap_tgt)/3);
		if (fuel_reserve < 0)fuel_reserve = 0;
		return true;
		}
}

bool SetBBTgt(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	if (!strnicmp(cbuf, "NULL", 4))
	{
		strcpy(refname,cbuf);
		afs_status = 0;
		return true;
	}
	OBJHANDLE bbtgt = oapiGetObjectByName(cbuf);
	if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, cbuf);
	if (bbtgt == NULL)return false;
	else 
		{
		strcpy(refname,cbuf);
		return true;
		}
}

bool SetOPE(void *id, char *str, void *data)
{
	float new_ope;
	sscanf(str, "%f", &new_ope);
	if (new_ope < 160)return false;
	else 
		{
			if (new_ope > op_tgt/1000)
			{
			op_meco_alt = op_tgt;
			op_tgt = new_ope*1000;
			} 
			else op_meco_alt = new_ope*1000;
			//ap_tgt = 0.67*op_meco_alt;
			if (((op_meco_alt*0.001)-200) > 0)
			{
			ap_tgt = 1000*((5.05*(sqrt((op_meco_alt*0.001)-200))) + 125);
			}
			else if (((op_meco_alt*0.001)-200) < 0)
			{
			ap_tgt = (op_meco_alt*0.625);
			}
			else ap_tgt = 125000;

			if (ap_tgt > (450000 - (3*fuel_reserve)))ap_tgt = (450000 - (3*fuel_reserve));
			if (ap_tgt < 100000)ap_tgt = 100000;
		return true;
		}
}

bool SetOAP(void *id, char *str, void *data)
{
	float new_oap;
	sscanf(str, "%f", &new_oap);
	if (new_oap < 160)return false;
	else 
		{
			if (new_oap < op_meco_alt/1000)
			{
			op_tgt = op_meco_alt;
			op_meco_alt = new_oap*1000;
			} 
			else op_tgt = new_oap*1000;
			//ap_tgt = 0.67*op_meco_alt;
			if (((op_meco_alt*0.001)-200) > 0)
			{
			ap_tgt = 1000*((5.05*(sqrt((op_meco_alt*0.001)-200))) + 125);
			}
			else if (((op_meco_alt*0.001)-200) < 0)
			{
			ap_tgt = (op_meco_alt*0.625);
			}
			else ap_tgt = 125000;

			if (ap_tgt > (450000 - (3*fuel_reserve)))ap_tgt = (450000 - (3*fuel_reserve));
			if (ap_tgt < 100000)ap_tgt = 100000;
		return true;
		}
}

bool SetPLD(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	OBJHANDLE hPld = oapiGetObjectByName(cbuf);
	if (hPld == NULL)return false;
	else
	{
	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hPld);
	DWORD ac = v->AttachmentCount(true);
	if (ac == 0) return false;
	for (DWORD j = 0; j < ac; j++)
		{
		ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle(true, j);
		OBJHANDLE hOp = v->GetAttachmentStatus(hAtt);
		if (hOp == NULL)
			{
			VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(hMe);
			if (vme->AttachChild(hPld, pld_attach, hAtt) == true)
				{
					pld_att_status = 1;
					calc_sw = 0;
					strcpy(pldname, cbuf);
					return true;
				}
			else return false;
			}
		else if (j == ac-1) return false;
		}
	
	}
	
}

BOOL CALLBACK DlgProc (HWND, UINT, WPARAM, LPARAM);

int f9_v12_stg2::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {

	HWND hDbox;

if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
	

	} 
  }

  else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		
	} 
  }

else { // unmodified keys
    switch (key) {

		//case OAPI_KEY_G: // remove fairings
		//return 1;

		//case OAPI_KEY_L: // descent pilot
		//RESETKEY (kstate, OAPI_KEY_L);
		//if (dlgl_status == 0)
		//{
		//double lat, lng, rds;
		//GetEquPos(lng,lat,rds);
		//min_inc = DEG*lat;
		//oapiOpenInputBox("Enter Equ.Inc(+North : -South)",SetINC, 0, 50, 0);
		//dlgl_status = 1;
		//}
		//else if (dlgl_status == 1)
		//{
		////oapiOpenInputBox("Enter Perigee Alt.> 160 (km)",SetOPE, 0, 50, 0);
		//oapiOpenInputBox("Enter Apogee Alt.> 160 (km)",SetOAP, 0, 50, 0);
		//dlgl_status = 2;
		//}
		//else if (dlgl_status == 2)
		//{
		////oapiOpenInputBox("Enter Apogee Alt.> 160 (km)",SetOAP, 0, 50, 0);
		//oapiOpenInputBox("Enter Perigee Alt.> 160 (km)",SetOPE, 0, 50, 0);
		//dlgl_status = 0;
		//}
		//return 1;
		
		

		case OAPI_KEY_J: // jettison pld
		if (intfair_status == 1)
		{
		hpd = GetAttachmentStatus(pld_attach);
		if (hpd == NULL);
		else 
			{
			DetachChild(pld_attach, 1);
			pld_att_status = 0;
			CalcEmptyMass();
			oapiSetFocusObject(hpd);
			strcpy(pldname, "NULL");
			}
		}
		else if (intfair_status == 0)
		{
		JettisonFair();
		}
		return 1;

		//case OAPI_KEY_P: // add payload
		//hMe = GetHandle();
		//if(pld_att_status == 0)oapiOpenInputBox("Enter Payload Name", SetPLD, 0, 50, 0);
		//return 1;

		//case OAPI_KEY_K:
		//hGref = GetGravityRef();
		//oapiOpenInputBox("Enter Flyback Target Name",SetBBTgt, 0, 50, 0);
		//return 1;

		//case OAPI_KEY_M:
		//if (dlg_status == 0)
		//{
		//oapiOpenInputBox("Enter Apogee at Stage1 MECO1 100-450 (km alt.)",SetAP, 0, 50, 0);
		//dlg_status = 1;
		//}
		//else if (dlg_status == 1)
		//{
		//oapiOpenInputBox("Enter Flyback Fuel Reserve 0-80000 (kg)",SetRF, 0, 50, 0);
		//dlg_status = 0;
		//}
		//return 1;

		case OAPI_KEY_V: // ascent pilot
		if (eng_status == 0)
			{
			if(ap_status == 0)ap_status = 1;
			else ap_status = 0;
			op_status = 0;
			KillAllRCS2();
			}
		else if (eng_status == 1)
			{
			if(op_status == 0)op_status = 1;
			else op_status = 0;
			ap_status = 0;
			KillAllRCS2();
			}
		apop_status = ap_status + op_status;
		return 1;

		case OAPI_KEY_E:
			if (cam_status == 0)cam_status = 1;
			else if (cam_status == 1)cam_status = 0;
			SetCamDir();
		return 1;

		case OAPI_KEY_K:
			hDbox = oapiFindDialog (g_Param.hDLL, IDD_CTRL);
			if (hDbox) oapiCloseDialog (hDbox);
			else hDbox = oapiOpenDialogEx(g_Param.hDLL, IDD_CTRL, DlgProc,DLG_CAPTIONCLOSE, this);
		return 1;

		case OAPI_KEY_B:
		if(GetThrusterGroupLevel(THGROUP_MAIN) == 0)
		{
			if(GetThrusterLevel(th_dump) == 0) SetThrusterLevel(th_dump, 1);
			else if(GetThrusterLevel(th_dump) > 0) SetThrusterLevel(th_dump, 0);
		}
		else SetThrusterLevel(th_dump, 0);
		return 1;

		case OAPI_KEY_N:
		if (eng_status == 0)
		{
			JettisonStg();
			if(ap_status == 1)
			{
				ap_status = 0;
				op_status = 1;
			}
		}
		return 1;

		case COMM_CALC_MASS:
		CalcEmptyMass();
		return 1;

		case COMM_CALC_FRE:
		CalcFuelRes();
		return 1;
	}
  }
  return 0;
}


DLLCLBK void InitModule (HINSTANCE hModule)
{
	g_Param.hDLL = hModule;
	oapiRegisterCustomControls (hModule);
	
	
}

DLLCLBK void ExitModule (HINSTANCE hModule)
{
	oapiUnregisterCustomControls (hModule);
	
}
void InitDialog(HWND hWnd)
{
	char cbuf[256];
	char lotbuf[2];
	char lotbuftemp[2];
	strcpy(lotbuftemp, "0");
	
	sprintf (cbuf, "%0.0f", fuel_reserve);
	SetWindowText (GetDlgItem (hWnd, IDC_FUEL_RES), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Enter");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_FUEL_APPLY), cbuf);

	sprintf (cbuf, "%s", refname);
	SetWindowText (GetDlgItem (hWnd, IDC_TGTNM), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Enter");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_TGTNM_APPLY), cbuf);

	if (afs_status == 0)sprintf(cbuf,"OFF");
	else sprintf(cbuf,"ENABLED");
	SetWindowText (GetDlgItem (hWnd, IDC_FB_STATUS), cbuf);
	if (eng_status == 0)sprintf (cbuf, "On/Off");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_FB_STATUS_APPLY), cbuf);

	sprintf (cbuf, "%s", pldname);
	SetWindowText (GetDlgItem (hWnd, IDC_PLD_NM), cbuf);
	if (pld_att_status == 0)SetWindowText (GetDlgItem (hWnd, IDC_PLD_NM_APPLY), "Attach");
	else SetWindowText (GetDlgItem (hWnd, IDC_PLD_NM_APPLY), "Jettison");

	sprintf (cbuf, "%0.0f kg", pld_mass);
	SetWindowText (GetDlgItem (hWnd, IDC_PLD_MASS), cbuf);

	sprintf (cbuf, "%0.2f", (tgt_inc*DEG));
	SetWindowText (GetDlgItem (hWnd, IDC_INC), cbuf);

	sprintf (cbuf, "%0.1f", (op_meco_alt*0.001));
	SetWindowText (GetDlgItem (hWnd, IDC_PER), cbuf);

	sprintf (cbuf, "%0.1f", (op_tgt*0.001));
	SetWindowText (GetDlgItem (hWnd, IDC_APO), cbuf);

	sprintf (cbuf, "%0.2f", (ap_tgt*0.001));
	SetWindowText (GetDlgItem (hWnd, IDC_APO1), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Enter");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_APO1_APPLY), cbuf);

	

	//////////////////
	if(lot_status == 1)
	{
		if (apop_status == 0)sprintf (cbuf, "ENABLED");
		else sprintf (cbuf, "ACTIVE");
	}
	else if (apop_status == 0)sprintf (cbuf, "OFF");
	else sprintf (cbuf, "ACTIVE");
	SetWindowText (GetDlgItem (hWnd, IDC_APOP), cbuf);

	/////////// launch timer
	sprintf (lotbuf, "%d", (lot_hh));
	if(lot_hh < 10) 
		{
		strcat(lotbuftemp,lotbuf);
		SetWindowText (GetDlgItem (hWnd, IDC_LOT_HH), lotbuftemp);
		strcpy(lotbuftemp, "0");
		}
	else SetWindowText (GetDlgItem (hWnd, IDC_LOT_HH), lotbuf);

	sprintf (lotbuf, "%d", (lot_mm));
	if(lot_mm < 10)
		{
		strcat(lotbuftemp,lotbuf);
		SetWindowText (GetDlgItem (hWnd, IDC_LOT_MM), lotbuftemp);
		strcpy(lotbuftemp, "0");
		}
	else SetWindowText (GetDlgItem (hWnd, IDC_LOT_MM), lotbuf);

	sprintf (lotbuf, "%d", (lot_ss));
	if(lot_ss < 10)
		{
		strcat(lotbuftemp,lotbuf);
		SetWindowText (GetDlgItem (hWnd, IDC_LOT_SS), lotbuftemp);
		strcpy(lotbuftemp, "0");
		}
	else SetWindowText (GetDlgItem (hWnd, IDC_LOT_SS), lotbuf);

	if(lot_status == 0)sprintf (cbuf, "OFF");
	else if(lot_status == 3)sprintf (cbuf, "OFF");
	else sprintf (cbuf, "ON");
	SetWindowText (GetDlgItem (hWnd, IDC_ALOT), cbuf);

	/////////////////////////////////////////

	if (intfair_status == 0)sprintf (cbuf, "Jettison");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_FAIR_APPLY), cbuf);

	if (eng_status == 0)sprintf (cbuf, "Jettison");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_JSTG_APPLY), cbuf);

	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
	v->SendBufferedKey(COMM_CALC_FRE);
	sprintf (cbuf, "%0.0f kg", fuel_res_est);
	SetWindowText (GetDlgItem (hWnd, IDC_FRE), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Apply");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_FRE_APPLY), cbuf);
	
	OBJHANDLE hSLC = v->GetAttachmentStatus(lpd_attach);
	if (hSLC)SetWindowText (GetDlgItem (hWnd, IDC_HDBK), "Apply");
	else SetWindowText (GetDlgItem (hWnd, IDC_HDBK), "N/A");

	if (edlo_sw == 0)sprintf (cbuf, "FLYBACK-EDL");
	else sprintf (cbuf, "EDL ONLY");
	SetWindowText (GetDlgItem (hWnd, IDC_EDLO), cbuf);
	if (eng_status == 0)sprintf (cbuf, "Change");
	else sprintf (cbuf, "N/A");
	SetWindowText (GetDlgItem (hWnd, IDC_EDLO_APPLY), cbuf);

	/////// asds pos
	sprintf (cbuf, "Lng. %0.3f  Lat. %0.3f", (asds_core_lng*DEG), (asds_core_lat*DEG));
	SetWindowText (GetDlgItem (hWnd, IDC_ASDS), cbuf);

	/////// frv pos
	sprintf (cbuf, "Lng. %0.3f  Lat. %0.3f", (frv_core_lng*DEG), (frv_core_lat*DEG));
	SetWindowText (GetDlgItem (hWnd, IDC_FRV), cbuf);
}

bool SetLOT(char *str1, char *str2, char *str3)
{
	float new_hh;
	sscanf(str1, "%f", &new_hh);
	float new_mm;
	sscanf(str2, "%f", &new_mm);
	float new_ss;
	sscanf(str3, "%f", &new_ss);

	if (new_hh > 23) return false;
	else if (new_hh < 0) return false;
	else if (new_mm > 59) return false;
	else if (new_mm < 0) return false;
	else if (new_ss > 59) return false;
	else if (new_ss < 0) return false;
	else
		{
			lot_hh = (int)(new_hh);
			lot_mm = (int)(new_mm);
			lot_ss = (int)(new_ss);
			return true;
		}
	return true;
}

BOOL CALLBACK DlgProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	
	

	switch (uMsg) {
	case WM_INITDIALOG:
		InitDialog(hWnd);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam)) 
		{
		char cbuf[256];
		char chh[256];
		char cmm[256];
		char css[256];

		case IDCANCEL:
			oapiCloseDialog (hWnd);
			return TRUE;
		case IDC_FUEL_APPLY:
			if (eng_status == 0)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_FUEL_RES), cbuf, 256);
			SetRF(0,cbuf,0);
			}
			InitDialog(hWnd);
			return TRUE;

		case IDC_FRE_APPLY:
			if (eng_status == 0)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_FRE), cbuf, 256);
			SetRF(0,cbuf,0);
			}
			InitDialog(hWnd);
			return TRUE;
		
		case IDC_TGTNM_APPLY:
			if (eng_status == 0)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_TGTNM), cbuf, 256);
			SetBBTgt(0,cbuf,0);
			}
			InitDialog(hWnd);
			return TRUE;

		case IDC_PLD_NM_APPLY:
			if (pld_att_status == 0)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_PLD_NM), cbuf, 256);
			SetPLD(0,cbuf,0);
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			v->SendBufferedKey(COMM_CALC_MASS);
			InitDialog(hWnd);
			}
			else if (pld_att_status == 1 && intfair_status == 1)
			{
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			v->SendBufferedKey(OAPI_KEY_J);
			InitDialog(hWnd);
			}
			return TRUE;
			
		case IDC_FB_STATUS_APPLY:
			if (eng_status == 0)
			{
			bchk = oapiGetObjectByName(refname);
			if (bchk == NULL)bchk = oapiGetBaseByName(hGref, refname);
			if (bchk == NULL)afs_status = 0;
			else if(afs_status == 0)afs_status = 1;
			else afs_status = 0;
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_INC_APPLY:
			GetWindowText (GetDlgItem (hWnd, IDC_INC), cbuf, 256);
			SetINC(0,cbuf,0);
			InitDialog(hWnd);
			return TRUE;

			case IDC_PER_APPLY:
			GetWindowText (GetDlgItem (hWnd, IDC_PER), cbuf, 256);
			SetOPE(0,cbuf,0);
			InitDialog(hWnd);
			return TRUE;

			case IDC_APO_APPLY:
			GetWindowText (GetDlgItem (hWnd, IDC_APO), cbuf, 256);
			SetOAP(0,cbuf,0);
			InitDialog(hWnd);
			return TRUE;

			case IDC_APO1_APPLY:
			if (eng_status == 0)
			{
			GetWindowText (GetDlgItem (hWnd, IDC_APO1), cbuf, 256);
			SetAP(0,cbuf,0);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_APOP_APPLY:
			{
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			v->SendBufferedKey(OAPI_KEY_V);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_FAIR_APPLY:
			if(intfair_status == 0)
			{
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			v->SendBufferedKey(OAPI_KEY_J);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_JSTG_APPLY:
			if(eng_status == 0)
			{
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			v->SendBufferedKey(OAPI_KEY_N);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_HDBK:
			{
			VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hMe);
			OBJHANDLE hLpad = v->GetAttachmentStatus(lpd_attach);
			if(hLpad) 
				{
				v = (VESSEL2*)oapiGetVesselInterface(hLpad);
				v->SendBufferedKey(OAPI_KEY_G);
				}
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_EDLO_APPLY:
			if(eng_status == 0)
			{
			if(edlo_sw == 0) edlo_sw = 1;
			else edlo_sw = 0;
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_LOT_APPLY:
			{
			GetWindowText (GetDlgItem (hWnd, IDC_LOT_HH), chh, 256);
			GetWindowText (GetDlgItem (hWnd, IDC_LOT_MM), cmm, 256);
			GetWindowText (GetDlgItem (hWnd, IDC_LOT_SS), css, 256);
			
			SetLOT(chh,cmm,css);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_ALOT_APPLY:
			{
			if(lot_status == 0)lot_status = 1;
			else if(lot_status == 1)lot_status = 0;
			oapiDelAnnotation (hNote);
			}
			InitDialog(hWnd);
			return TRUE;

			case IDC_REFRESH:
			InitDialog(hWnd);
			return TRUE;

		}
		
		break;
	}
	return oapiDefDialogProc (hWnd, uMsg, wParam, lParam);
}


void f9_v12_stg2::clbkPreStep(double SimT, double SimDT, double MJD)
{
////////// launch timer
if(lot_status == 1)DoLOT();
else if(lot_status == 2)DoLOT();

///////// max q trail
double dynp = GetDynPressure();
if(dynp > 22000)
	{
	pst_lvl = (dynp-22000)/3000;
	if (pst_lvl > 1)pst_lvl = 1;
	if(eng_status > 0)pst_lvl = 0;
	}
else pst_lvl = 0;


////////////////////////// l pad condensation
if (GetAttachmentStatus(lpd_attach) == NULL) pst_lvl2 = 0;
else if (oapiGetTimeAcceleration() > 10) pst_lvl2 = 0;
else pst_lvl2 = 0.7;

step_time = SimDT;///////
double simdt = SimDT;

double mth_lev = GetThrusterLevel(th_main[eng_status]);
if(eng_status == 1)
{
if (mth_lev == 0 && rcs_sw2 == 0) ConfigRCS2();
else if (mth_lev > 0 && rcs_sw2 == 1) RevertRCS2();
}

if(mth_lev > 0)SetThrusterLevel(th_dump, 0);

if (calc_sw == 0)
{
CalcEmptyMass();
calc_sw = 1;
}

if (ap_status == 1) DoAscent();
else if (op_status == 1) DoOrbit();
apop_status = op_status+ap_status;
////////////////engine light
if (mth_lev > 0 && eng_status == 0)
	{
	double rndm = 0.15*(oapiRand());
	rndm = (mth_lev*0.9) - rndm;
	if (rndm < 0)rndm = 0;
	else if (rndm > 1)rndm = 1;
	spt_light->Activate(true);
	spt_light->SetIntensity(rndm);

	}
	else spt_light->Activate(false);
//////////

}



void f9_v12_stg2::clbkPostStep (double simt, double simdt, double mjd)
{
	
	
}




bool f9_v12_stg2::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);

	//int chw = 16;

	
	char test[256];
	strcpy(test,"abcdefghijklmnopqrstuvwxyz");
	int it = skp->GetTextWidth(test, 26);
	it = (int)(it/26);
	int chw = it;

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	//int w10 = (int)(344*d);
	//int w11 = (int)(360*d);
	//int w12 = (int)(376*d);
	//int w13 = (int)(392*d);
	//int w14 = (int)(408*d);
	//int w15 = (int)(424*d);


	

	if (eng_status == 1)
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	double m1 = GetMass();
	double m2 = 0;
	if (eng_status == 0)m2 = GetPropellantMass(hProp1);
	else m2 = GetPropellantMass(hProp2);
	double test = (m1/(m1-m2));
	double isp_curr = GetThrusterIsp(th_main[eng_status]);
	double zdv = (isp_curr*(log(test)));

	sprintf(abuf,"Remaining dV = %0.2f m/s : RCS Prop. = %0.1fkg", zdv, GetPropellantMass(hRCSProp2));
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	double m1 = GetMass();
	double m2 = 0;
	if (eng_status == 0)m2 = GetPropellantMass(hProp1);
	else m2 = GetPropellantMass(hProp2);
	double test = (m1/(m1-m2));
	double isp_curr = GetThrusterIsp(th_main[eng_status]);
	double zdv = (isp_curr*(log(test)));

	sprintf(abuf,"Remaining dV = %0.2f m/s", zdv);
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"Payload: %s %dkg", pldname, int(pld_mass));
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"Tgt.Orb: Inc.%0.2f ApA.%dkm PeA.%dkm",abs(tgt_inc*DEG),int(op_tgt/1000),int(op_meco_alt/1000));
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	if (eng_status == 0)
	{
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"1st Stg: Fuel Res.%dkg ApA.%dkm",int(fuel_reserve),int(ap_tgt/1000));
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	if (afs_status == 0)
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(edlo_sw == 1)sprintf(abuf,"1st Stg: EDL ONLY OFF: TGT. %s", refname);
	else sprintf(abuf,"1st Stg: FLYBACK-EDL OFF: TGT. %s", refname);
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	if(edlo_sw == 1)sprintf(abuf,"1st Stg: EDL ONLY ON: Tgt. %s", refname);
	else sprintf(abuf,"1st Stg: FLYBACK-EDL ON: Tgt. %s", refname);
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"K = Launch Control Panel On/Off");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}

	if (intfair_status == 0)
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"J = Jettison Fairing");
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}
	else if (pld_att_status == 1)
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"J = Jettison Payload");
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}

	if (eng_status == 0)
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"N = Jettison 1st Stage");
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}

	if ((ap_status + op_status) > 0)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Launch Autopilot Active");
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"V = Start Launch Autopilot");
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"E = Set Camera +Z/-Z");
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"B = 2nd Stage Fuel Dump");
	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
	}

	return true;
}

void f9_v12_stg2::RevertRCS2 (void)
{
	SetThrusterGroupLevel(htgrppup2,0);
	SetThrusterGroupLevel(htgrppdown2,0);
	SetThrusterGroupLevel(htgrpyright2,0);
	SetThrusterGroupLevel(htgrpyleft2,0);

	DelThrusterGroup(htgrppup2,THGROUP_ATT_PITCHUP, false);
	DelThrusterGroup(htgrppdown2,THGROUP_ATT_PITCHDOWN, false);
	DelThrusterGroup(htgrpyright2,THGROUP_ATT_YAWRIGHT, false);
	DelThrusterGroup(htgrpyleft2,THGROUP_ATT_YAWLEFT, false);

	th_group[0] = th_rcs[14];
	htgrppup2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[15];
	htgrppdown2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[13];
	htgrpyright2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[12];
	htgrpyleft2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWLEFT);

	rcs_sw2 = 0;
}

void f9_v12_stg2::ConfigRCS2 (void)
{
	SetThrusterGroupLevel(htgrppup2,0);
	SetThrusterGroupLevel(htgrppdown2,0);
	SetThrusterGroupLevel(htgrpyright2,0);
	SetThrusterGroupLevel(htgrpyleft2,0);

	DelThrusterGroup(htgrppup2,THGROUP_ATT_PITCHUP, false);
	DelThrusterGroup(htgrppdown2,THGROUP_ATT_PITCHDOWN, false);
	DelThrusterGroup(htgrpyright2,THGROUP_ATT_YAWRIGHT, false);
	DelThrusterGroup(htgrpyleft2,THGROUP_ATT_YAWLEFT, false);

	th_group[0] = th_rcs[16];
	th_group[1] = th_rcs[17];
	htgrppup2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[18];
	th_group[1] = th_rcs[19];
	htgrppdown2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[20];
	th_group[1] = th_rcs[22];
	htgrpyright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[21];
	th_group[1] = th_rcs[23];
	htgrpyleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	rcs_sw2 = 1;
}

void f9_v12_stg2::JettisonFair (void)
{
VESSELSTATUS vs;
 GetStatus(vs); // get mother vessel status vs
  VECTOR3 rvel;
  Local2Rel (STG1_FAIR1_OFF, vs.rpos); // child vessel position absolute in global frame
  if (eng_status == 1)Local2Rel (STG2_FAIR1_OFF, vs.rpos); // ignore this! (different pos if 1st stage already jettisoned)
  GlobalRot (_V(0,-4,0), rvel);  // child vessel velocity relative to mother, convert to global frame
  vs.rvel = vs.rvel + rvel;  // child vessel velocity absolute in global frame
  vs.eng_main = vs.eng_hovr = 0.0; // engines off
 vs.status = 0; // in flight
  vs.vrot.x = vs.vrot.x - 0.2;	// child vessel rotation relative to mother
  vs.vrot.z = vs.vrot.z;	// child vessel rotation relative to mother
  vs.vrot.y = vs.vrot.y; // child vessel rotation relative to mother
 /* char myname[256];
  oapiGetObjectName(hMe,myname,30);
  char fairname[256];
  strcpy(fairname, myname);
  strcat(fairname, "_Fair1");*/
  if (disable_fair_rec == 0) oapiCreateVessel ("Fair1", "Falcon9_BLOCK4/f9_v12_fair_1", vs);
  else oapiCreateVessel ("Fair1", "Falcon9_BLOCK4/f9_v12_fair_nopara", vs);

  GetStatus(vs);
  Local2Rel (STG1_FAIR2_OFF, vs.rpos);
  if (eng_status == 1)Local2Rel (STG2_FAIR2_OFF, vs.rpos);
  GlobalRot (_V(0,4,0), rvel);
  vs.rvel = vs.rvel + rvel;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  vs.vrot.x = -(vs.vrot.x + 0.2);
  vs.vrot.z = vs.vrot.z;
  vs.vrot.y = -vs.vrot.y;

	VECTOR3 xright;
	VECTOR3 yup;
	VECTOR3 zforward;

	GlobalRot(_V(-1,0,0), xright);
	GlobalRot(_V(0,-1,0), yup);
	GlobalRot(_V(0,0, 1), zforward);

	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);
	
	vs.arot = _V(alpha, beta, (PI/2)-gamma);

 /* strcpy(fairname, myname);
  strcat(fairname, "_Fair2");*/
  if (disable_fair_rec == 0) oapiCreateVessel ("Fair2", "Falcon9_BLOCK4/f9_v12_fair_2", vs);
  else oapiCreateVessel ("Fair2", "Falcon9_BLOCK4/f9_v12_fair_nopara", vs);

  if (eng_status == 1)
  {
  DelMesh(4);
  DelMesh(3);
  DelMesh(2);
  DelMesh(1);
  }
  else
  {
  DelMesh(4);
  DelMesh(3);
  }

intfair_status = 1;
CalcEmptyMass();

PlayVesselWave(MyID, SEP, 0, 200, 30000);
PlayVesselWave(MyID, SEPFR, 0, 225, 0);
}


void f9_v12_stg2::JettisonStg (void)
{
  VESSELSTATUS vs;
  //char name[256];
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (STG1_OFF, vs.rpos);
  GlobalRot (_V(0,0,-5), rvel);
  vs.rvel = vs.rvel + rvel;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  vs.vrot.x = vs.vrot.x;
  vs.vrot.z = vs.vrot.z;
  vs.vrot.y = vs.vrot.y;
  char myname[256];
  oapiGetObjectName(hMe,myname,30);
  char stgname[256];
  strcpy(stgname, myname);
  strcat(stgname, "_Stg1");
  OBJHANDLE hStg1;
  if(nolegs == 0) hStg1 = oapiCreateVessel (stgname, "Falcon9_BLOCK4/f9_v12_stg1", vs);
  else hStg1 = oapiCreateVessel (stgname, "Falcon9_BLOCK4/f9_v12_stg1_expendable", vs);
  double fuel_trans = GetPropellantMass(hProp1);
  VESSEL2* v =(VESSEL2*)oapiGetVesselInterface(hStg1);
  //VESSEL *v = oapiGetVesselInterface (hStg1);
  DWORD nAttach = v->GetPropellantCount ();
  for (DWORD j = 0; j < nAttach; j++)
	{
	PROPELLANT_HANDLE ph = v->GetPropellantHandleByIndex (j);
	if(j == 0)v->SetPropellantMass(ph,fuel_trans);
	else
		{
		double maxfuel = v->GetPropellantMaxMass(ph);
		v->SetPropellantMass(ph,maxfuel);
		}
	}
  v->clbkConsumeBufferedKey(COMM_SEND_STR,false,refname);
  if(afs_status == 1)
  {
	v->clbkConsumeBufferedKey(COMM_SEND_AFS,false,refname);
	if(edlo_sw == 1)v->clbkConsumeBufferedKey(COMM_SEND_EDLO,false,refname);
  }

  VESSELSTATUS2 vss;
  memset (&vss, 0, sizeof(vss));
  vss.version = 2;
  GetStatusEx(&vss);
  Local2Rel (STG1STG2_OFF, vss.rpos);
  DefSetStateEx (&vss);

	DelMesh(2);
	DelMesh(1);
	eng_status = 1;


	CalcEmptyMass();
	ConfigMain();

	PlayVesselWave(MyID, SEP, 0, 200, 25000);
	PlayVesselWave(MyID, SEPST, 0, 225, 0);

}

void f9_v12_stg2::ConfigMain (void)
{
	KillAllRCS2();
	SetThrusterLevel(th_main[0],0);
	DelThrusterGroup(htgrpmain,THGROUP_MAIN, false);
	DelThrusterGroup(htgrpbleft2,THGROUP_ATT_BANKLEFT, false);
	DelThrusterGroup(htgrpbright2,THGROUP_ATT_BANKRIGHT, false);
	DelThrusterGroup(htgrppup2,THGROUP_ATT_PITCHUP, false);
	DelThrusterGroup(htgrppdown2,THGROUP_ATT_PITCHDOWN, false);
	DelThrusterGroup(htgrpyleft2,THGROUP_ATT_YAWLEFT, false);
	DelThrusterGroup(htgrpyright2,THGROUP_ATT_YAWRIGHT, false);

	//DelPropellantResource(hProp1);
	SetPropellantMass(hProp1, 0);
	SetDefaultPropellantResource(hProp2);

	th_group[0] = th_rcs[14];
	htgrppup2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[15];
	htgrppdown2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[13];
	htgrpyright2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[12];
	htgrpyleft2 = CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[16];
	th_group[1] = th_rcs[19];
	htgrpbright2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[17];
	th_group[1] = th_rcs[18];
	htgrpbleft2 = CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);
	
	th_group[0] = th_main[1];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);


	SetAttachmentParams(pld_attach,_V( 0, 0, 7.05), _V( 0,0,1), _V(0,1,0));

	for (UINT i = 0; i < GetMeshCount(); i++)
	{
	ShiftMesh(i,-STG1STG2_OFF);
	}

	SetSize (10);
	SetPMI(_V(18,18,1.9));
	SetCrossSections (_V(44,44,10));
	SetTouchdownPoints ( tdvtx_gearup, 4);
	ClearAirfoilDefinitions ();
	CreateAirfoil (LIFT_VERTICAL, _V(0,0,0), VLiftCoeff, 5, 20, 0.16);
	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeff, 5, 20, 0.16);
}

void f9_v12_stg2::KillAllRCS2()
{
SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
}

void f9_v12_stg2::CalcEmptyMass(void)
{
	double totalmass = STG2DRYMASS;
	
	if (eng_status == 0) totalmass = totalmass +  STG1DRYMASS;
	if (intfair_status == 0) totalmass = totalmass +  FAIR_MASS;

	OBJHANDLE hpld = GetAttachmentStatus(pld_attach);
	if (hpld == NULL)
	{
	SetEmptyMass(totalmass);
	pld_att_status = 0;
	pld_mass = 0;
	}
	else
	{
		pld_mass = oapiGetMass(hpld);
		totalmass = totalmass + pld_mass;
		oapiGetObjectName(hpld, pldname, 30);
		SetEmptyMass(totalmass);
		pld_att_status = 1;
	}

}
void f9_v12_stg2::CalcFuelRes(void)
{
	//////// check if landed and tgt valid
	VESSELSTATUS vs;
	GetStatus(vs);
	if (vs.status == 0) 
	{
		fuel_res_est = 0;
		return;
	}
	OBJHANDLE bbtgt = oapiGetObjectByName(refname);
	if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, refname);
	if (bbtgt == NULL)
	{		
		fuel_res_est = 0;
		CalcAsdsPos();
		return;
	}
	///////////////
	VECTOR3 tgt_gpos = _V(0,0,0);
	oapiGetGlobalPos(bbtgt,&tgt_gpos);
	VECTOR3 tgt_lpos = _V(0,0,0);
	Global2Local(tgt_gpos, tgt_lpos);
	VECTOR3 tgt_hpos = _V(0,0,0);
	HorizonRot(tgt_lpos, tgt_hpos);
	tgt_hpos.y = 0;

	

	double apo_factor = (250000 - ap_tgt);
	if (apo_factor < 0)apo_factor = 0;
	apo_factor = (apo_factor*3.5)+250000;

	double azm_angle = CalcAzimuth();
	VECTOR3 imp_hpos = _V(0,0,0);
	imp_hpos.x = sin(azm_angle)*apo_factor;
	imp_hpos.z = cos(azm_angle)*apo_factor;

	double tgt_imp_dist = dist(imp_hpos,tgt_hpos);
	VECTOR3 imp_mid_pnt = _V((imp_hpos.x*0.25),0,(imp_hpos.z*0.25));
	VECTOR3 imp_mid_to_tgt = tgt_hpos - imp_mid_pnt;
	normalise(imp_mid_to_tgt);
	normalise(imp_mid_pnt);
	double cos_dl = dotp(imp_mid_to_tgt, imp_mid_pnt);

	double dl_factor = -cos_dl*2000;
	
	if (edlo_sw == 0)
	{
	fuel_res_est =  26500 + (tgt_imp_dist*0.01*3) + dl_factor;
	if (fuel_res_est < 29000)fuel_res_est = 29000;
	}
	else
	{
	fuel_res_est = 23000;
	//fuel_res_est =  21500 + (tgt_imp_dist*0.01*0.5);
	//if (fuel_res_est < 22000)fuel_res_est = 22000;
	}
	if (fuel_res_est > 80000)fuel_res_est = 80000;
	//sprintf(oapiDebugString(), "fre %0.1f tid %0.1f apf %0.1f", fuel_res_est, tgt_imp_dist, apo_factor);

	CalcAsdsPos();

}
void f9_v12_stg2::CalcAsdsPos(void)
{
// stg1 ap at meco in km
	double algo_ap = ap_tgt * 0.001;
	double algo_ap2 = algo_ap * algo_ap;

	double core_range_15600 = 1000*((-2.092883971E-2*algo_ap2) + (6.11044315*algo_ap) + 158.2077267);
	double core_range_6000 = 1000*((-1.639211208E-2*algo_ap2) + (5.162726067*algo_ap) + 255.785993);

	double delta_range_kg = (core_range_15600 - core_range_6000)/9600;
	double pld_15600_diff = 15600 - pld_mass;
	double core_range = core_range_15600 - (delta_range_kg*pld_15600_diff);

	OBJHANDLE hEarth = GetGravityRef();
	VECTOR3 falcon_pos = _V(0,0,0);
	GetRelativePos(hEarth,falcon_pos);
	VECTOR3 falcon_pos_nm = falcon_pos;
	normalise(falcon_pos_nm);

	double azm_angle = CalcAzimuth();
	double algo_range = core_range*0.001;
	double algo_range2 = algo_range*algo_range;
	double azm_range_factor = RAD*0.01*((8.654755941E-3*algo_ap2) - (2.433711805*algo_ap) + 258.983414);//(0.15*RAD) + (RAD*algo_ap*0.006);
	azm_angle = azm_angle + azm_range_factor;

	VECTOR3 vel_vector_hz = _V((sin(azm_angle)),0,(cos(azm_angle)));
	VECTOR3 vel_vector_hz_nm = vel_vector_hz;
	normalise(vel_vector_hz_nm);

	VECTOR3 vel_vector_loc = _V(0,0,0);
	HorizonInvRot(vel_vector_hz, vel_vector_loc);
	VECTOR3 vel_vector_loc_nm = vel_vector_loc;
	normalise(vel_vector_loc_nm);

	VECTOR3 vel_vector = _V(0,0,0);
	GlobalRot(vel_vector_loc, vel_vector);
	VECTOR3 vel_vector_nm = vel_vector;
	normalise(vel_vector_nm);

	VECTOR3 rot_axis = crossp(falcon_pos_nm, vel_vector_nm);
	VECTOR3 rot_axis_nm = rot_axis;
	normalise(rot_axis_nm);

	double core_rot_angle = (core_range*0.935)/rt;
	if(edlo_sw == 0) core_rot_angle = (core_range*0.65)/rt;

	double rot_angle = core_rot_angle;
	VECTOR3 rot_vec = rot_axis_nm;

	MATRIX3 RG = rotm(rot_vec,rot_angle);
	VECTOR3 core_asds_vect = mul(RG, falcon_pos);


	double frf = ((38700-fuel_reserve)/5200)*0.08;
	RG = rotm(rot_vec,(rot_angle*1.2));
	if(edlo_sw == 0) RG = rotm(rot_vec,(rot_angle*(1.4+frf)));
	if(fuel_reserve == 0) RG = rotm(rot_vec,(rot_angle*2.3));
	VECTOR3 core_fvr_vect = mul(RG, falcon_pos);

	double asds_rad = 0;

	VECTOR3 earth_pos = _V(0,0,0);
	oapiGetGlobalPos(hEarth,&earth_pos);
	core_asds_vect = core_asds_vect+earth_pos;
	oapiGlobalToEqu(hEarth,core_asds_vect, &asds_core_lng, &asds_core_lat, &asds_rad);
	if(fuel_reserve == 0)
	{
		asds_core_lng = 0;
		asds_core_lat = 0;
	}

	double fvr_rad = 0;
	core_fvr_vect = core_fvr_vect+earth_pos;
	oapiGlobalToEqu(hEarth,core_fvr_vect, &frv_core_lng, &frv_core_lat, &fvr_rad);
	if(intfair_status == 1)
	{
		frv_core_lng = 0;
		frv_core_lat = 0;
	}
	if(disable_fair_rec == 1)
	{
		frv_core_lng = 0;
		frv_core_lat = 0;
	}
	//sprintf(oapiDebugString(), "asdscore lng %0.3f lat %0.3f core_range %0.3f azm_rng_factor %0.3f", (asds_core_lng*DEG), (asds_core_lat*DEG),core_range*0.001,azm_range_factor*DEG);
	//sprintf(oapiDebugString(), "fvr_core lng %0.3f lat %0.3f frf %0.3f", (frv_core_lng*DEG), (frv_core_lat*DEG), frf);
}

void f9_v12_stg2::DoLOT(void)
{
	oapiDelAnnotation (hNote);
	HWND hDlg;
	double lot_secs = (3600*lot_hh) + (60*lot_mm) + (lot_ss);
	double current_mjd = oapiGetSimMJD();
	double ct_secs_frac = current_mjd - (int)(current_mjd);
	double ct_secs = 86400*ct_secs_frac;
	//double ct_secs_diff = ct_secs - (int)(ct_secs);

	int ct_secs_int = (int)(ct_secs);
	//if(ct_secs_diff > 0) ct_secs_int = ct_secs_int + 1;

	int lot_secs_int = (int) (lot_secs);
	if (lot_secs_int < ct_secs_int)lot_secs_int = lot_secs_int + 86400;
	

	int t_minus = (lot_secs_int - ct_secs_int);
	
	int t_minus_hh = (int)(t_minus/3600);
	int t_minus_mm = (int)((t_minus - (t_minus_hh * 3600))/60);
	int t_minus_ss = (int)(t_minus - (t_minus_hh * 3600) - (t_minus_mm * 60));

char cbuf_hh[2];
char cbuf_mm[2];
char cbuf_ss[2];
char cbuf[256];

if (abs(t_minus_hh) < 10)sprintf(cbuf_hh, "0%d", t_minus_hh);
else sprintf(cbuf_hh, "%d", t_minus_hh);

if (abs(t_minus_mm) < 10)sprintf(cbuf_mm, "0%d", t_minus_mm);
else sprintf(cbuf_mm, "%d", t_minus_mm);

if (abs(t_minus_ss) < 10)sprintf(cbuf_ss, "0%d", t_minus_ss);
else sprintf(cbuf_ss, "%d", t_minus_ss);

sprintf(cbuf,"T - %s:%s:%s", cbuf_hh,cbuf_mm,cbuf_ss);

VECTOR3 col = _V(1,0.9,0);
hNote = oapiCreateAnnotation(true, 3, col);
oapiAnnotationSetPos(hNote,0.36,0.89,0.8,0.99);
oapiAnnotationSetText(hNote, cbuf);

	if (lot_status == 1)
	{
	if ((lot_secs_int - ct_secs_int) < 0) return;
	else if ((lot_secs_int - ct_secs_int) < 11) 
		{
		SendBufferedKey(OAPI_KEY_V);
		lot_status = 2;
		hDlg = oapiFindDialog (g_Param.hDLL, IDD_CTRL);
		if (hDlg) InitDialog(hDlg);
		}
	}
	if (lot_status == 2)
	{
	lot_timer = lot_timer + step_time;
	if (lot_timer > 10)
		{
		sprintf(cbuf,"T - 00:00:00");
		oapiAnnotationSetText(hNote, cbuf);
		}
	if (lot_timer > 11) 
		{
		lot_status = 3;
		lot_timer = 0;
		oapiDelAnnotation (hNote);
		hDlg = oapiFindDialog (g_Param.hDLL, IDD_CTRL);
		if (hDlg) InitDialog(hDlg);
		return;
		}
	}

}

void f9_v12_stg2::DoAscent(void)
{
if (eng_status == 1)return;


CalcAzimuth();
CalcFlightVector();

double th_max = GetThrusterMax(th_main[eng_status]);
double mass = GetMass();
double maxacc = th_max/mass;
double acc_lim = 39.24;
if (drg == 1)acc_lim = 29.43;
double englev = acc_lim/maxacc;
if (englev > 1)englev = 1;


if(ap_function == 0)
{
if (tu_sw == 0)
{
OBJHANDLE hLpad = GetAttachmentStatus(lpd_attach);
if (!hLpad) tu_sw = 1;
else tu_sw = 2;
}

if (tu_sw == 1)
	{
	if (timer == 0) PlayVesselWave(MyID, CTDN, 0, 200, 0);
	timer = timer+step_time;
	if (timer > 10) 
		{
		englev = 1;
		timer = 10;
		}
	else englev = 0;
	}
else if (tu_sw == 2)
	{
	if (timer == 0) PlayVesselWave(MyID, CTDN, 0, 200, 0);
	timer = timer+step_time;
	if (timer > 10) timer = 10;
	if (timer < 10) englev = (timer-8)/2;
	if (englev < 0)englev = 0;
	}
}
SetThrusterGroupLevel(THGROUP_MAIN,englev);

if (GetAltitude(ALTMODE_GROUND) > 70 && ap_function == 0)
{
	ap_function = 1;
	timer = 0;
}
if (ap_function > 0) RollToAz();
//if (ap_function == 0) GoToFlightVector();
if (ap_function == 1) GoToFlightVector();
if (ap_function == 2) GoToFlightVector();
if (ap_function == 3) GoToFlightVector();
if (ap_function == 4) GoToFlightVector();
if (ap_function == 5)
	{
	KillAllRCS2();
	SetThrusterGroupLevel(THGROUP_MAIN,0);
	ActivateNavmode(NAVMODE_KILLROT);
	ap_function = 0;
	ap_status = 0;
	JettisonStg();
	op_status = 1;
	}




}

double f9_v12_stg2::CalcAzimuth()
{
	double lon,lat,rad;
	GetEquPos(lon,lat,rad);
	if(abs(tgt_inc)<abs(lat)){
		tgt_inc=abs(lat);
	}else{
		tgt_inc=tgt_inc;
	}
	/////////////
	OBJHANDLE hearth=GetSurfaceRef();
	VECTOR3 relvel = _V(0,0,0);
	VECTOR3 globpos = _V(0,0,0);
	GetRelativeVel(hearth,relvel);
	GetGlobalPos(globpos);
	VECTOR3 relvel_glob = relvel+globpos;
	VECTOR3 rel_vel_loc = _V(0,0,0);
	Global2Local(relvel_glob,rel_vel_loc);

	VECTOR3 ov_hrz = _V(0,0,0);
	HorizonRot(rel_vel_loc, ov_hrz);

	double orb_ground_spd = sqrt((ov_hrz.x*ov_hrz.x)+(ov_hrz.z*ov_hrz.z));

	VECTOR3 orb_ground_vector = _V(ov_hrz.x,0,ov_hrz.z);

	////////
	double tgt_inc_sc = tgt_inc;
	if (GetAltitude(ALTMODE_GROUND) > 10000) // bloody winds at launch!!
	{
		if(lat > 0)
		{
		if (ov_hrz.z < -1 && tgt_inc > 0) tgt_inc_sc = -tgt_inc;
		}
		else if(lat < 0)
		{
		if (ov_hrz.z > 1 && tgt_inc < 0) tgt_inc_sc = -tgt_inc;
		}
	}
	double final_azimuth = asin(cos(tgt_inc_sc)/cos(lat));
	final_azimuth+=(90*RAD-final_azimuth)*(1-(tgt_inc_sc/abs(tgt_inc_sc)));
	VECTOR3 final_ground_vector = _V((FinalV()*sin(final_azimuth)),0,(FinalV()*cos(final_azimuth)));

	VECTOR3 req_ground_vector = final_ground_vector - orb_ground_vector;

	azimuth = atan2(req_ground_vector.x,req_ground_vector.z);

	/////////////
	//azimuth=atan2(FinalV()*(sin(asin(cos(tgt_inc)/cos(lat))))-GetOS(),FinalV()*(cos(asin(cos(tgt_inc)/cos(lat)))));
	//azimuth+=(90*RAD-azimuth)*(1-(tgt_inc/abs(tgt_inc)));
	
	vazim1=_V(sin(azimuth),0,cos(azimuth));
	
	HorizonInvRot(vazim1,azim_local);
	//sprintf(oapiDebugString(), "ap_function %d", ap_function);
	return azimuth;
}
double f9_v12_stg2::GetOS(){
	OBJHANDLE hearth=GetSurfaceRef();
	VECTOR3 relvel;
	GetRelativeVel(hearth,relvel);
	double os=length(relvel);
	return os;
}

double f9_v12_stg2::FinalV()
{
//double tgtapo = 180000;
//double tgtperi = -6318000;
double tgtapo = 200000;
double tgtperi = 2000000;
double abside=tgtapo;
	if(tgtapo<=260000){
			abside=tgtapo;
		}else{
			abside=tgtperi;}
if (op_status == 1) return sqrt(mu*(2/(op_meco_alt+rt)-1/(0.5*(op_tgt+op_meco_alt+(2*rt)))));
else return sqrt(mu*(2/(abside+rt)-1/(0.5*(tgtapo+tgtperi+2*rt))));

}

void f9_v12_stg2::CalcFlightVector()
{
	double initial_pitch = RAD*(77.2+(5*((ap_tgt - 125000)*0.00001))); //77.5

	double fuel_res_fctr = sin(PI05*fuel_reserve/100000);
	if (fuel_res_fctr < 0)fuel_res_fctr = 0;
	else if (fuel_res_fctr > 1)fuel_res_fctr = 1;
	fuel_res_fctr=3.5*fuel_res_fctr*RAD; // 3

	initial_pitch = initial_pitch + fuel_res_fctr;

	double azmth = CalcAzimuth();
	double azmth_fctr = RAD*(-0.25*(sin(azmth)));

	initial_pitch = initial_pitch + azmth_fctr;

	if (ap_function == 2) 
	{
		VECTOR3 haspdv = _V(0,1,0);
		GetGroundspeedVector (FRAME_HORIZON, haspdv);
		double mass = GetMass();
		VECTOR3 thrustvector = _V(0,0,0);
		GetThrustVector(thrustvector);
		double trst = length(thrustvector);
		double acc = trst/mass;
		double time_avail = (200-length(haspdv))/(acc-9.8);
		double pitch_rate = ((pitch_angle)-initial_pitch)/time_avail;
		pitch_angle = pitch_angle - (step_time*pitch_rate);

		//sprintf(oapiDebugString(), "initial_pitch %0.4f pitch_angle %0.4f", initial_pitch*DEG,pitch_angle*DEG);

		if (length(haspdv) > 200)ap_function = 3;


	}
	
	
	else if (ap_function == 3)
	{
		VECTOR3 haspdv = _V(0,1,0);
		GetGroundspeedVector (FRAME_HORIZON, haspdv);
		double xzlngth = sqrt((haspdv.x*haspdv.x)+(haspdv.z*haspdv.z));
		if (haspdv.y < 0)haspdv.y = 0;
		pitch_angle = atan2(haspdv.y,xzlngth);
		if (GetAltitude() > 30000) ap_function = 4;
		//if (length(haspdv) < 200) sprintf(oapiDebugString(), "gspd ptch angle %0.4f ", pitch_angle*DEG);
	}

	else if (ap_function == 4)
	{
		double ap_dist = 0;
		GetApDist(ap_dist);

		double englev = GetThrusterGroupLevel(THGROUP_MAIN);
		double fuel_remaining = GetPropellantMass(hProp1) - fuel_reserve;
		double isp_curr = GetThrusterIsp(th_main[eng_status]);
		double th_max = GetThrusterMax(th_main[eng_status]);
		double fuel_rate = englev*th_max/isp_curr;
		double mintime = fuel_remaining/fuel_rate;
		if (mintime < 0)mintime = 0;

		/////////////////////////

		VECTOR3 haspdv = _V(0,1,0);
		GetGroundspeedVector (FRAME_HORIZON, haspdv);
		//double xzlngth = sqrt((haspdv.x*haspdv.x)+(haspdv.z*haspdv.z));
		if (haspdv.y < 0)haspdv.y = 0;

		///////////////////////////////////////////////////////
		double vmass = GetMass();

		OBJHANDLE hearth=GetSurfaceRef();
		VECTOR3 relvel = _V(0,0,0);
		GetRelativeVel(hearth,relvel);
		double orb_vel = length(relvel);
		double orb_grnd_spd = sqrt((orb_vel*orb_vel)-(haspdv.y*haspdv.y));
		double circ = PI2*(rt+GetAltitude());
		double ang_vel = PI2*orb_grnd_spd/circ;
		VECTOR3 angmomvect = _V(0,(-1*ang_vel),0);
		VECTOR3 velvect = _V(0,0,orb_grnd_spd);
		VECTOR3 corvect = crossp(angmomvect,velvect);
		double acc_cor = length(corvect);

		double sinp = sin(GetPitch());
		double dv_grnd = isp_curr*log(vmass/(vmass-fuel_remaining));
		dv_grnd = dv_grnd*sinp;

		double final_orb_grnd_spd = orb_grnd_spd+dv_grnd;
		velvect = _V(0,0,final_orb_grnd_spd);
		VECTOR3 final_corvect = crossp(angmomvect,velvect);
		double final_acc_cor = length(corvect);

		VECTOR3 weight = _V(0,0,0);
		GetWeightVector(weight);
		
		double gravt = length(weight);
		double gg = (gravt/vmass)-final_acc_cor;
		double aa = (mintime*mintime)/(2*gg);
		double bb = (haspdv.y*mintime/gg)+((mintime*mintime)/2);
		double cc = -(ap_tgt - GetAltitude() - (haspdv.y*mintime) - ((haspdv.y*haspdv.y)/(2*gg)) );

		double req_acc_1 = (-bb+(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);
		double req_acc_2 = (-bb-(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);

		//if (mintime < 4) englev = englev-(step_time/5);
		//if(englev < 0.1)englev = 0.1;

		double curr_acc = th_max*englev/(vmass-(fuel_remaining*0.5));
		
		//if (ap_dist > (rt+ap_tgt))ap_function = 5;
		if (fuel_remaining < (0.5*fuel_rate*step_time))ap_function = 5;
		
		double total_req_acc = req_acc_1+gg-acc_cor;

		double psc = total_req_acc/curr_acc;
		//if (psc > 0.985)psc = 0.985;
		if (psc > 0.9962)psc = 0.9962;
		else if (psc < -0.9962)psc = -0.9962;
		double pangle = asin(psc);

		if (mintime < 2)pitch_angle = pitch_angle;
		else pitch_angle = pangle;
		
		//sprintf(oapiDebugString(), "af %d bias %0.4f reqaprt %0.2f aprate %0.2f", ap_function, bias, req_ap_rate, ap_rate);
		//sprintf(oapiDebugString(), "pangle %0.4f ", pangle*DEG);
		
		
	}

		
		vazim2 = _V(vazim1.x, (tan(pitch_angle)), vazim1.z);

	//if (ap_function < 2)vazim2 = _V(0,1,0);
	if (ap_function < 2)
	{
	VECTOR3 haspdv = _V(0,1,0);
	GetGroundspeedVector (FRAME_HORIZON, haspdv);
	vazim2 = _V(-haspdv.x*0.5,haspdv.y,-haspdv.z*0.5);
	}
	HorizonInvRot(vazim2,flight_vector);

		//sprintf(oapiDebugString(), "af %d", ap_function);
}

void f9_v12_stg2::RollToAz()
{
DeactivateNavmode(NAVMODE_KILLROT);
double roll_angle = atan2(azim_local.x,-azim_local.y);
if (roll_angle > PI) roll_angle = roll_angle - PI2;
else if (roll_angle < -PI) roll_angle = roll_angle + PI2;

	double mass = GetMass();

	double max_rcs = GetThrusterMax(th_rcs[0]);
	double max_torque = 2*max_rcs*4;
	double inertia = 1.9*mass;

	if (eng_status == 1)
	{
	max_rcs = GetThrusterMax(th_rcs[19]);
	max_torque = 2*max_rcs*((1.706*0.866)-(0.123*0.25));
	inertia = 1.9*mass;
	}
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);


	double max_acc = max_torque/inertia;

	double max_drot = max_acc*step_time;

	double reqzv = 0.9*(sqrt(abs(roll_angle*max_acc*2)));
	if (roll_angle > 0)reqzv = -reqzv;

	double req_drot = reqzv - avel.z;

	double thz_lvl = req_drot/max_drot;


	if((roll_angle*DEG) < 1 && (roll_angle*DEG) > -1)
	{
	reqzv = 1*-roll_angle;
	double acc_reqz = (reqzv-avel.z)/0.1;
	thz_lvl = acc_reqz/max_acc;
	}


	if(thz_lvl > 1)thz_lvl = 1;
	else if(thz_lvl < -1)thz_lvl = -1;


	if(thz_lvl > 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, thz_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}
	else if(thz_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, -thz_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}

	if (ap_function == 1)
	{

		if (roll_angle > -0.05 && roll_angle < 0.05)
		{
		VECTOR3 aspdv = _V(0,1,0);
		GetHorizonAirspeedVector(aspdv);
		double azimuth = CalcAzimuth();
		double az_factor = 30*(1 - sin(azimuth));
		if (aspdv.y > (55)) //+az_factor
			{
			vrot_lim = 2*RAD;
			ap_function = 2;
			}
		}
	}
}



void f9_v12_stg2::GoToFlightVector()
{
	//VECTOR3 wv = _V(0,0,0);
	//GetWeightVector(wv);
	//wv = -wv;

	
	//double angoa = atan2(-wv.y, wv.z);
	//double slipage = atan2(-wv.x, wv.z);
	DeactivateNavmode(NAVMODE_KILLROT);

	double angoa = atan2(-flight_vector.y, flight_vector.z);
	double slipage = atan2(flight_vector.x, flight_vector.z);

	if(angoa > PI)angoa = (PI*2)-angoa;
	else if(angoa < -PI)angoa = (PI*2)+angoa;

	if(slipage > PI)slipage = (PI*2)-slipage;
	else if(slipage < -PI)slipage = (PI*2)+slipage;

	double mass = GetMass();

	double max_rcs = GetThrusterMax(th_rcs[0]);
	double max_torque = 2*max_rcs*21;
	double inertia = 129*mass;
	
	if (eng_status == 1)
	{
	max_rcs = GetThrusterMax(th_rcs[13]);
	max_torque = 1*max_rcs*9.8;
	inertia = 18*mass;
	if (rcs_sw2 == 1)
		{
		max_rcs = GetThrusterMax(th_rcs[16]);
		max_torque = 2*max_rcs*((4.365*0.866)-(0.123*0.433));
		}
	}

	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);
	double max_acc = max_torque/inertia;
	double max_drot = max_acc*step_time;

	double reqxv = 0.9*(sqrt(abs(angoa*max_acc*2)));
	if (reqxv > vrot_lim)reqxv = vrot_lim;
	if (angoa > 0)reqxv = -reqxv;
	double req_drotx = reqxv - avel.x;
	double thx_lvl = req_drotx/max_drot;
	if((angoa*DEG) < 1 && (angoa*DEG) > -1)
	{
	reqxv = 1*-angoa;
	double acc_reqx = (reqxv-avel.x)/0.1;
	thx_lvl = acc_reqx/max_acc;
	}

	double reqyv = 0.9*(sqrt(abs(slipage*max_acc*2)));
	if (reqyv > vrot_lim)reqyv = vrot_lim;
	if (slipage > 0)reqyv = -reqyv;
	double req_droty = reqyv - avel.y;
	double thy_lvl = req_droty/max_drot;
	if((slipage*DEG) < 1 && (slipage*DEG) > -1)
	{
	reqyv = 1*-slipage;
	double acc_reqy = (reqyv-avel.y)/0.1;
	thy_lvl = acc_reqy/max_acc;
	}

	//sprintf(oapiDebugString(), "slip %0.4f yrot %0.4f reqyv%0.4f", (slipage*DEG),(avel.y*DEG), (reqyv*DEG));


	if(thx_lvl > 1)thx_lvl = 1;
	else if(thx_lvl < -1)thx_lvl = -1;

	if(thy_lvl > 1)thy_lvl = 1;
	else if(thy_lvl < -1)thy_lvl = -1;


	if(thx_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		}
	else if(thx_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
	}


	if(thy_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, thy_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		}
	else if(thy_lvl < 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -thy_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
	}


}

bool f9_v12_stg2::clbkLoadGenericCockpit ()
{
	SetCamDir();
return true;
}

void f9_v12_stg2::SetCamDir()
{
	VECTOR3 camdefdir = _V(0,0,1);
	if(cam_status == 1)camdefdir = _V(0,0,-1);
	SetCameraDefaultDirection(camdefdir, 0);
	oapiCameraAttach(GetHandle(), 2);
}

void f9_v12_stg2::DoOrbit()
{
	
	
	if (start_sw == 0)
	{
	timer = timer+step_time;
	if (timer < 5) return;
	else
		{
		start_sw = 1;
		timer = 0;
		SetThrusterLevel(th_main[1],1);
		}
	}

	else if (start_sw == 1 && intfair_status == 0)
	{
	timer = timer+step_time;
	if (timer > 20)
		{
			JettisonFair();
			timer = 0;
		}
	}


DeactivateNavmode(NAVMODE_KILLROT);
CalcAzimuth();

double fnl_vel = sqrt(mu*(2/(op_meco_alt+rt)-1/(0.5*(op_tgt+op_meco_alt+(2*rt)))));

double emp_mass = GetEmptyMass();
double vmass = GetMass();

OBJHANDLE hearth=GetSurfaceRef();

VECTOR3 haspdv = _V(0,0,0);
GetGroundspeedVector (FRAME_HORIZON, haspdv);
	
VECTOR3 weight = _V(0,0,0);
GetWeightVector(weight);

//centrifugal acc
	VECTOR3 relvel = _V(0,0,0);
	GetRelativeVel(hearth,relvel);
	double orb_vel = length(relvel);
	double orb_grnd_spd = sqrt((orb_vel*orb_vel)-(haspdv.y*haspdv.y));
	double rds = (rt + GetAltitude());
	double ang_vel = orb_grnd_spd/rds;
	double acc_cent_fact = 1;
	double acc_cent = ang_vel*ang_vel*rds*acc_cent_fact;
		

	double gravt = length(weight);
	double g = -(gravt/vmass)+acc_cent;

	double aa = g/2;
	double bb = haspdv.y;
	double cc = GetAltitude()-op_meco_alt;

	//double timetoimp = (-bb-(sqrt((bb*bb)-(4*aa*cc))))/(2*aa);

	double max_th = GetThrusterMax(th_main[1]);
	double max_acc = max_th/vmass;
	double acc_lim = 50;
	if (drg > 0) acc_lim = 30;
	double max_eng_lvl = 1;
	if (max_acc > acc_lim) max_eng_lvl = acc_lim/max_acc;
	if (max_eng_lvl < 0.7)max_eng_lvl = 0.7;
	if ((fnl_vel-orb_vel) < 400)
	{
	if (max_eng_lvl > 0.7)max_eng_lvl = 0.7;
	}
	double sinp = sin(GetPitch());
	double cosp = cos(GetPitch());
	
	double dv_req = fnl_vel - orb_grnd_spd;
	double fraction = exp(dv_req/GetThrusterIsp(th_main[1]));
	double fuelmass = vmass - (vmass/fraction);

	double dth = step_time * 0.2;

	double englevel = GetThrusterGroupLevel(THGROUP_MAIN);
	
	double fuelrate = GetThrusterMax(th_main[1])*englevel/GetThrusterIsp(th_main[1]);
	double dvtime = fuelmass/fuelrate;
	dvtime = dvtime/cosp;

	GetElements (hearth,el,&prm,0,FRAME_EQU);
	double ap_time = prm.ApT;
	//sprintf(oapiDebugString(), "dvtime %0.1f ap_time %0.1f", dvtime, ap_time);
	double ap_time_limiter = dvtime/ap_time;
	if (ap_time_limiter > 1)ap_time_limiter = 1;
	else if (ap_time_limiter < 0.7)ap_time_limiter = 0.7;
	if (ap_time_limiter < max_eng_lvl) max_eng_lvl = ap_time_limiter;
	

	double vvel_fnl = 0;

	//double acc_req = 2*((vvel_fnl*dvtime)+GetAltitude()-op_meco_alt)/(dvtime*dvtime);
	//double acc_req = 2*2*(op_meco_alt-GetAltitude()-(haspdv.y*dvtime))/(dvtime*dvtime);
	double acc_req = ((6*(op_meco_alt-GetAltitude() - (haspdv.y*dvtime)))+(2*dvtime*haspdv.y))/(dvtime*dvtime);
	
	acc_req = acc_req - g;

	//sprintf(oapiDebugString(), "acc_req %0.1f", acc_req);

	double srp = acc_req/(max_acc*englevel);
	if (srp > 1)srp = 1;
	else if (srp < -1)srp = -1;
	double req_pitch = asin(srp);
	double crp = cos(req_pitch);

	VECTOR3 fv2d = _V(0, srp, crp);
	VECTOR3 fv_hrz = _V(0,0,0);
	fv_hrz.x = sin(azimuth)*crp;
	fv_hrz.y = srp;
	fv_hrz.z = cos(azimuth)*crp;

	double max_dv = step_time*max_acc*englevel;
	double req_englevel = (fnl_vel-orb_vel)/max_dv;
	if (req_englevel < 0)
	{
		op_status = 0;
		englevel = 0;
		KillAllRCS2();
		SetThrusterLevel(th_main[1],0);
		ActivateNavmode(NAVMODE_KILLROT);
		return;
	}
	//englevel = (fnl_vel-orb_vel)/max_dv;
	if (req_englevel > max_eng_lvl)req_englevel = max_eng_lvl;
	

	if (englevel > req_englevel)
	{
		englevel = englevel-dth;
		if (englevel < req_englevel) englevel = req_englevel;
	}
	else if (englevel < req_englevel)
	{
		englevel = englevel+dth;
		if (englevel > req_englevel) englevel = req_englevel;
	}
	

	if ((fnl_vel-orb_vel) < 30)
	{
		//flight_vector = flight_vector;
		ActivateNavmode(NAVMODE_KILLROT);
		///if(englevel > 0)englevel = 0.7;
		///if (englevel > max_eng_lvl)englevel = max_eng_lvl;
		//sprintf(oapiDebugString(), "fv-ov %0.1f", (fnl_vel-orb_vel));
	}

	//else if ((fnl_vel-orb_vel) < 400)
	//{
	//	if(englevel > 0)englevel = 0.7;
	//	if (englevel > max_eng_lvl)englevel = max_eng_lvl;
	//	HorizonInvRot(fv_hrz, flight_vector);
	//	GoToFlightVector();
	//}
	else 
	{
	HorizonInvRot(fv_hrz, flight_vector);
	GoToFlightVector();
	}

	if (start_sw == 0)englevel = 0;
	SetThrusterLevel(th_main[1],englevel);




}